/*
 *  Copyright (c) 2006           Ji YongGang <jungle@soforge.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/***************************************************************************
 *          extract_chmLib.c - CHM archive extractor                       *
 *                           -------------------                           *
 *                                                                         *
 *  author:     Jed Wing <jedwin@ugcs.caltech.edu>                         *
 *  notes:      This is a quick-and-dirty chm archive extractor.           *
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2.1 of the  *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <chm_lib.h>

#include "chmsee.h"
#include "utils.h"
#include "exchm.h"

static int dir_exists(const char *);
static int rmkdir(char *);
static int _extract_callback(struct chmFile *, struct chmUnitInfo *, void *);

struct extract_context
{
        const char *base_path;
        const char *hhc_file;
        const char *hhk_file;
        const char *codeset;
};

static int 
dir_exists(const char *path)
{
        struct stat statbuf;
        if (stat(path, &statbuf) != -1)
                return 1;
        else
                return 0;
}

static int 
rmkdir(char *path)
{
        /*
         * strip off trailing components unless we can stat the directory, or we
         * have run out of components
         */

        char *i = rindex(path, '/');

        if(path[0] == '\0'  ||  dir_exists(path))
                return 0;

        if (i != NULL)
        {
                *i = '\0';
                rmkdir(path);
                *i = '/';
                mkdir(path, 0777);
        }

        if (dir_exists(path))
                return 0;
        else
                return -1;
}

/*
 * callback function for enumerate API
 */
static int
_extract_callback(struct chmFile *h, struct chmUnitInfo *ui, void *context)
{
        unsigned char buffer[32768];
        struct extract_context *ctx = (struct extract_context *)context;
        char *i;
        gchar *filename_utf8;

        if (ui->path[0] != '/') {
                return CHM_ENUMERATOR_CONTINUE;
        }

        LOG_DEBUG("ui->path = %s", ui->path);

        /* encoding of extracted filename is UTF8 */
        filename_utf8 = convert_filename_to_utf8(ui->path, ctx->codeset);

        if (snprintf(buffer, sizeof(buffer), "%s%s", ctx->base_path, filename_utf8) > 1024)
                return CHM_ENUMERATOR_FAILURE;

        if (ui->length != 0) {
                FILE *fout;
                LONGINT64 len, remain=ui->length;
                LONGUINT64 offset = 0;

                gchar *file_ext;

                file_ext = g_strrstr(g_path_get_basename(filename_utf8), ".");
                LOG_DEBUG("file_ext = %s", file_ext);

                /* find out hhc and hhk files */
                if (file_ext) {
                        if (ncase_compare_utf8_string(".hhc", file_ext) == 0) {
                                ctx->hhc_file = g_strdup_printf("%s%s", ctx->base_path, filename_utf8);
                        }
                        if (ncase_compare_utf8_string(".hhk", file_ext) == 0) {
                                ctx->hhk_file = g_strdup_printf("%s%s", ctx->base_path, filename_utf8);
                        }
                }

                if ((fout = fopen(buffer, "wb")) == NULL) {
                        /* make sure that it isn't just a missing directory before we abort */ 
                        unsigned char newbuf[32768];
                        strcpy(newbuf, buffer);
                        i = rindex(newbuf, '/');
                        *i = '\0';
                        rmkdir(newbuf);
                        if ((fout = fopen(buffer, "wb")) == NULL)
                                return CHM_ENUMERATOR_FAILURE;
                }

                while (remain != 0) {
                        len = chm_retrieve_object(h, ui, buffer, offset, 32768);
                        if (len > 0) {
                                fwrite(buffer, 1, (size_t)len, fout);
                                offset += len;
                                remain -= len;
                        } else {
                                break;
                        }
                }

                fclose(fout);

        } else {
                if (rmkdir(buffer) == -1)
                        return CHM_ENUMERATOR_FAILURE;
        }

        g_free(filename_utf8);

        return CHM_ENUMERATOR_CONTINUE;
}

int
ex_chm_file(ChmSee *chmsee, const gchar *filename) 
{
        struct chmFile *h;
        struct extract_context ec;
        h = chm_open(filename);

        if (h == NULL) {
                return 1;
        }

        ec.base_path = (const char *) chmsee->book_dir;
        ec.codeset = (const char *) chmsee->current_codeset;
        ec.hhc_file = NULL;
        ec.hhk_file = NULL;

        if (! chm_enumerate(h, CHM_ENUMERATE_ALL, _extract_callback, (void *)&ec)) {
                return 1;
        }

        chm_close(h);
        chmsee->hhc_name = ec.hhc_file;
        chmsee->hhk_name = ec.hhk_file;

        return 0;
}
