/*
 *  Copyright (c) 2006           Ji YongGang <jungle@soforge.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gtk/gtk.h>
#include <libxml/parser.h>
#include <libxml/HTMLparser.h>

#include "chmsee.h"
#include "parse_hh.h"
#include "utils.h"

static gint depth = 0;
static gint prev_depth = 0;
static gboolean tree_item = FALSE;
static gchar *title = NULL;
static gchar *local = NULL;

static GtkTreeModel *model;
static GtkTreeIter parent_iter;
static GtkTreeIter last_iter;

static void startDocumentHH(void *);
static void endDocumentHH(void *);
static void startElementHH(void *, const xmlChar *, const xmlChar **);
static void endElementHH(void *, const xmlChar *);


static xmlSAXHandler hhSAXHandlerStruct = {
    NULL, /* internalSubset */
    NULL, /* isStandalone */
    NULL, /* hasInternalSubset */
    NULL, /* hasExternalSubset */
    NULL, /* resolveEntity */
    NULL, /* getEntity */
    NULL, /* entityDecl */
    NULL, /* notationDecl */
    NULL, /* attributeDecl */
    NULL, /* elementDecl */
    NULL, /* unparsedEntityDecl */
    NULL, /* setDocumentLocator */
    startDocumentHH, /* startDocument */
    endDocumentHH, /* endDocument */
    startElementHH, /* startElement */
    endElementHH, /* endElement */
    NULL, /* reference */
    NULL, /* characters */
    NULL, /* ignorableWhitespace */
    NULL, /* processingInstruction */
    NULL, /* comment */
    NULL, /* xmlParserWarning */
    NULL, /* xmlParserError */
    NULL, /* xmlParserError */
    NULL, /* getParameterEntity */
    NULL, /* cdataBlock */
    NULL, /* externalSubset */
    1,    /* initialized */
    NULL, /* private */
    NULL, /* startElementNsSAX2Func */
    NULL, /* endElementNsSAX2Func */
    NULL  /* xmlStructuredErrorFunc */
};

static xmlSAXHandlerPtr hhSAXHandler = &hhSAXHandlerStruct;

static void
startDocumentHH(void *ctx)
{
        ChmSee *chmsee = (ChmSee *)ctx;

        LOG_INFO("SAX.startDocument()");

        model = gtk_tree_view_get_model(GTK_TREE_VIEW (chmsee->chm_tree));
        g_object_ref(model);
        gtk_tree_view_set_model(GTK_TREE_VIEW(chmsee->chm_tree), NULL);
}

static void
endDocumentHH(void *ctx)
{
        ChmSee *chmsee = (ChmSee *)ctx;

        LOG_INFO("SAX.endDocument()");
        
        gtk_tree_view_set_model(GTK_TREE_VIEW(chmsee->chm_tree), model);
        g_object_unref(model);
}

static void
startElementHH(void *ctx, const xmlChar *name, const xmlChar **atts)
{
        LOG_DEBUG("SAX.startElement(%s)", (char *)name);
        
        if (g_ascii_strcasecmp("ul", (char *)name) == 0) {
                depth++;
        } else if (g_ascii_strcasecmp("object", (char *)name) == 0) {
                guchar *type;

                type = g_strdup(atts[1]);

                LOG_DEBUG("type = %s", type);

                if (g_ascii_strcasecmp("text/sitemap", type) == 0) {
                        tree_item = TRUE;
                }

                g_free(type);
        } else if (g_ascii_strcasecmp("param", (char *)name) == 0) {
                guchar *param_name;
                guchar *param_value;

                param_name = g_strdup(atts[1]);
                param_value = g_strdup(atts[3]);

                if (tree_item) {
                        if (g_ascii_strcasecmp("Name", param_name) == 0) {
                                title = g_strdup(param_value);
                        } else if (g_ascii_strcasecmp("Local", param_name) == 0) {
                                local = g_strdup(param_value);
                        }
                }
         
                g_free(param_name);
                g_free(param_value);
        }
}

static void
endElementHH(void *ctx, const xmlChar *name)
{
        ChmSee *chmsee = (ChmSee *)ctx;

        LOG_DEBUG("SAX.endElement(%s)", (char *) name);

        if (g_ascii_strcasecmp("ul", (char *)name) == 0) {
                depth--;
        } else if (g_ascii_strcasecmp("object", (char *)name) == 0) {
                if (tree_item) {
                        GtkTreeIter iter;
                        gchar *local_lowcase = NULL;

                        tree_item = FALSE;

                        LOG_DEBUG("title = %s", title);
                        LOG_DEBUG("local = %s", local);

                        if (local) {
                                local_lowcase = g_utf8_strdown(local, -1);
                                g_free(local);
                                local = NULL;
                        } else {
                                local_lowcase = g_strdup(CHMSEE_NO_LINK);
                        }

                        LOG_DEBUG("prev_depth = %d", prev_depth);
                        LOG_DEBUG("depth = %d", depth);

                        if (depth == 1) {
                                gtk_tree_store_append(GTK_TREE_STORE (model), &iter, NULL);
                                                
                                if (!chmsee->home_url)
                                        chmsee->home_url = g_strdup(local_lowcase);
                        } else {
                                if (depth > prev_depth) {
                                        parent_iter = last_iter;
                                } else if (depth < prev_depth) {
                                        gtk_tree_model_iter_parent(model, &iter, &parent_iter);
                                        parent_iter = iter;
                                }
                                
                                gtk_tree_store_append(GTK_TREE_STORE (model), &iter, &parent_iter);
                        }

                        gtk_tree_store_set(GTK_TREE_STORE (model),
                                           &iter,
                                           TITLE_COLUMN, title,
                                           FILENAME_COLUMN, local_lowcase,
                                           -1);

                        last_iter = iter;
                        prev_depth = depth;

                        g_free(title);
                        g_free(local_lowcase);
                }
        }
}

int
parse_hhc_file(ChmSee *chmsee)
{
        htmlDocPtr doc = NULL;
        
        doc = htmlSAXParseFile(chmsee->hhc_name, chmsee->current_codeset, hhSAXHandler, chmsee);

        if (doc != NULL) {
            LOG_INFO("htmlSAXParseFile returned non-NULL");
            xmlFreeDoc(doc);
        }

        return 0;
}
